/*
CRC32 -- Calculate 16 and 32 bit CRCs for one or more files
    Copyright (C) 2003  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    For Windows: Compile with Visual Studio 6 or later.  
    Requires comdlg32.lib odbc32.lib odbccp32.lib winmm.lib
	cl crc32.c setargv 

    For Linux: Compile with gcc:  gcc crc32.c

    The CRC code was taken from various sources.  I've copied in 
    the comments from that code.

*/

/*

The following C code (by Rob Warnock <rpw3@sgi.com>) does CRC-32 in
BigEndian/BigEndian byte/bit order.  That is, the data is sent most
significant byte first, and each of the bits within a byte is sent most
significant bit first, as in FDDI. You will need to twiddle with it to do
Ethernet CRC, i.e., BigEndian/LittleEndian byte/bit order. [Left as an
exercise for the reader.]

The CRCs this code generates agree with the vendor-supplied Verilog models
of several of the popular FDDI "MAC" chips.

*/

/*
 * Copyright (C) 1986 Gary S. Brown.  You may use this program, or
 * code or tables extracted from it, as desired without restriction.
 */

/* First, the polynomial itself and its table of feedback terms.  The  */
/* polynomial is                                                       */
/* X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0 */
/* Note that we take it "backwards" and put the highest-order term in  */
/* the lowest-order bit.  The X^32 term is "implied"; the LSB is the   */
/* X^31 term, etc.  The X^0 term (usually shown as "+1") results in    */
/* the MSB being 1.                                                    */

/* Note that the usual hardware shift register implementation, which   */
/* is what we're using (we're merely optimizing it by doing eight-bit  */
/* chunks at a time) shifts bits into the lowest-order term.  In our   */
/* implementation, that means shifting towards the right.  Why do we   */
/* do it this way?  Because the calculated CRC must be transmitted in  */
/* order from highest-order term to lowest-order term.  UARTs transmit */
/* characters in order from LSB to MSB.  By storing the CRC this way,  */
/* we hand it to the UART in the order low-byte to high-byte; the UART */
/* sends each low-bit to hight-bit; and the result is transmission bit */
/* by bit from highest- to lowest-order term without requiring any bit */
/* shuffling on our part.  Reception works similarly.                  */

/* The feedback terms table consists of 256, 32-bit entries.  Notes:   */
/*                                                                     */
/*     The table can be generated at runtime if desired; code to do so */
/*     is shown later.  It might not be obvious, but the feedback      */
/*     terms simply represent the results of eight shift/xor opera-    */
/*     tions for all combinations of data and CRC register values.     */
/*                                                                     */
/*     The values must be right-shifted by eight bits by the "updcrc"  */
/*     logic; the shift must be unsigned (bring in zeroes).  On some   */
/*     hardware you could probably optimize the shift in assembler by  */
/*     using byte-swap instructions.                                   */



#define BUFSI 60000
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#include <time.h>
#include <malloc.h>
#include <ctype.h>


#ifdef _WIN32
#include <process.h>
#include <io.h>
#include <direct.h>
#define DELIM '\\'
#define DELIMSTR "\\"
#else
#include <limits.h>
#define _MAX_DIR PATH_MAX 
#define _MAX_PATH PATH_MAX 
#define O_BINARY 0
#define DELIM '/'
#define DELIMSTR "/"
#endif

char *buffer,prefix[_MAX_DIR],finm[_MAX_PATH],current_path[_MAX_DIR];

#ifndef _WIN32
char *strupr(char *s) {
  char c;
  while (*s++) {
    c = *s;
    *s = toupper(c);
  }
}

int _chdrive(int c) {
  return 0;
}



#endif

int main(int,char *[]);
unsigned short updcrc(unsigned short,unsigned char *,unsigned);


char filenm[_MAX_PATH];
int getfilename(int,char *[],char*);
unsigned numdrives;

#define CRC32_POLY 0x04c11db7     /* AUTODIN II, Ethernet, & FDDI */
unsigned long crc32_table[256];

unsigned long
updcrc32(unsigned long icrc, unsigned char *icp, unsigned int icnt) {
	register unsigned long crc = icrc;
	register unsigned char *cp = icp;
	register unsigned int cnt = icnt;

	while ( cnt-- ) {
		crc = (crc<<8) ^ crc32_table[(crc>>24) ^ *cp++];
	}

	return( crc );
}

void init_crc32()
{
        int i, j;
        unsigned long c;

        for (i = 0; i < 256; ++i) {
                for (c = i << 24, j = 8; j > 0; --j)
                        c = c & 0x80000000 ? (c << 1) ^ CRC32_POLY : (c << 1);
                crc32_table[i] = c;
        }
}


int main(int argc,char *argv[]) {
	int i, infi;
	struct stat statbuf;
	long filesize;
	char *pa;
	unsigned l;
	unsigned short crc16;
	unsigned long crc32;

	if (argc < 2 || argv[1][1] == '?') {
		fprintf(stderr,"CRC32 %s\n",__DATE__);
		fprintf(stderr,"\nCopywrite 1995 - 2003, Norm London, London Computing.\nnorm_london@londoncomputing.com\n");
		fprintf(stderr,"Portions Copyright (C) 1986 Gary S. Brown.\n");
		fprintf(stderr,"Portions written by Rob Warnock\n");
        	fprintf(stderr,"This program is free software; you can redistribute it and/or modify\n");
		fprintf(stderr,"it under the terms of the GNU General Public License as published by\n");
		fprintf(stderr,"the Free Software Foundation; either version 2 of the License, or\n");
		fprintf(stderr,"(at your option) any later version.\n\n");
		fprintf(stderr,"This program is distributed in the hope that it will be useful,\n");
		fprintf(stderr,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
		fprintf(stderr,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
		fprintf(stderr,"GNU General Public License for more details.\n\n");
		fprintf(stderr,"You should have received a copy of the GNU General Public License\n");
		fprintf(stderr,"along with this program; if not, write to the Free Software\n");
		fprintf(stderr,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");
		fprintf(stderr,"Usage is: CRC32 @responsefile | file1 file2 ...\n");
		fprintf(stderr,"Calculates 16 and 32 bit Cyclic Redundancy Check Codes.\n");
		fprintf(stderr,"@responsefile -- responsefile is the name of a file containing\n");
		fprintf(stderr,"filenames.  Wildcards not permitted. @stdin, read from stdin.\n");
		fprintf(stderr,"Wildcards permitted on command line only.\n");
		fprintf(stderr,"The following information is displayed for each file:\n");
		fprintf(stderr,"16 Bit CRC,32 Bit CRC,File Size,File Modification Time(Integer),\n");
		fprintf(stderr,"File Modification Time(Text),File Name,Path Name.\n");
		exit(1);
	}
	init_crc32();	

	if ((buffer = (char *)malloc((unsigned)(BUFSI))) == NULL) {
		fprintf(stderr,"Insufficient Memory\n");
		exit(4);
	}

	getcwd(current_path,sizeof(current_path));

	while (getfilename(argc,argv,filenm)) {
		/* for each argument, extract prefix and look up in catalog */
		strcpy(prefix,filenm);
#ifdef _WIN32
		// Parse out filename and directory assuming dos/windows conventions
		for (i=strlen(prefix)-1;i && prefix[i]!='\\' && prefix[i]!=':' && prefix[i] != '/'; --i) ;
		if (prefix[i]==':' || prefix[i]=='\\') {
			strcpy(finm,prefix+i+1);
			if (prefix[i] == '\\')
				if (i == 0 || (i == 2 && prefix[1] == ':'))
					prefix[++i] = 0;
				else
					prefix[i] = 0;
			else
				prefix[++i] = 0;

			if (prefix[0] == '\\' && prefix[1] == '\\')
                		strupr(prefix);

			else {
				if (prefix[1] == ':' && prefix[0] != current_path[0])
					_chdrive(toupper(prefix[0])-'@');
				chdir(prefix);
				pa = getcwd(prefix,sizeof(prefix));
				if (prefix[0] != current_path[0])
					_chdrive(toupper(current_path[0])-'@');
				chdir(current_path);
			}
		}
		else {
			strcpy(prefix,current_path);
			strcpy(finm,filenm);
		}
#else
		// Parse out filename and directory assuming UNIX conventions
		for (i=strlen(prefix)-1;i && prefix[i] != '/';--i) ;
		if (prefix[i]=='/') {
			strcpy(finm,prefix+i+1);
			// break string into two pieces, one for the path, one for the filename
			if (prefix[i] == '/')
				if (i == 0)
					prefix[++i] = 0;
				else
					prefix[i] = 0;
			else
				prefix[++i] = 0;

			if (prefix[0] == '/' && prefix[1] == '/')
                ;
			else {
				chdir(prefix);
				pa = getcwd(prefix,sizeof(prefix));
				chdir(current_path);
			}
		}
		else {
			strcpy(prefix,current_path);
			strcpy(finm,filenm);
		}
#endif


		/* created an ambiguous filename to search directory */
 		/* returns an unambiguous filename to open */
		if (stat(filenm,&statbuf)) {
			perror(filenm);
			continue;
		}
		if (statbuf.st_mode & S_IFDIR)
			continue;
		if ((infi = open(filenm,O_RDONLY+O_BINARY)) < 1) {
			fprintf(stderr,"Unable to open ");
			perror(filenm);
			continue;
		}

		filesize = 0;
		crc16 = (unsigned) (0xFFFF);	/* use an sdlc like crc by setting initial */
		crc32 = (unsigned) (0xFFFFFFFF);
										/* fstat(infi,&statbuf); */
		if (statbuf.st_size > 0) {
			do {
				l = read(infi,buffer,(unsigned)(BUFSI));
				filesize += l;
				crc16 = updcrc(crc16,buffer,l);
				crc32 = updcrc32(crc32,buffer,l);
			}
			while (l == BUFSI);
		}
		strcpy(buffer,ctime(&statbuf.st_mtime));
		if (buffer[(l=strlen(buffer)-1)] == '\n')
			buffer[l] = 0;
		if (prefix[strlen(prefix)-1] != DELIM && prefix[strlen(prefix)-1] != '/')
			strcat(prefix,DELIMSTR);
		strcat(prefix,finm);
		printf("%4x,%8x,%9ld,%9ld,%s,%-16s,%s\n",
		crc16,crc32,filesize,statbuf.st_mtime,buffer,finm,prefix);
		close(infi);
	}
	exit(0);
}

FILE *redir;

int getfilename(int argc,char *argv[],char *filenm) {
	static int k = 1;
	static int fileopen = 0;
	int more,l;

	more = 1;
	for (;;) {
		if (fileopen) {
			if (fgets(filenm,_MAX_PATH,redir) == NULL) {
				fileopen = 0;
				fclose(redir);
			}
			else {
				l = strlen(filenm);
				if (filenm[l-1] == '\n')
					filenm[l-1] = 0;	
				return 1; /* got a filename */
			}
		}

		if (k >= argc)
			return 0; /* done */

		if (argv[k][0] == '@') {
			strcpy(filenm,argv[k]+1);
			strupr(filenm);
			if (strcmp(filenm,"STDIN") == 0)
				redir = stdin;
			else
				redir = fopen(argv[k]+1,"r");
			if (redir == NULL)
				fprintf(stderr,"Unable to open response file %s\n",argv[k]+1);
			else
				fileopen = 1;
			++k;
		}
		else {
			strcpy(filenm, argv[k++]);
			return 1;
		}
	}
}

/* Calculate, intelligently, the CRC of a dataset incrementally given a
 * buffer full at a time.
 * Initialize crc to 0 for XMODEM, -1 for CCITT.
 *
 * Usage:
 *       newcrc = updcrc( oldcrc, bufadr, buflen )
 *           unsigned int oldcrc, buflen;
 *           char *bufadr;
 *
 * Compile with -DTEST to generate program that prints CRC of stdin to stdout.
 * Compile with -DMAKETAB to print values for crctab to stdout
 */

/* the CRC polynomial. This is used by XMODEM (almost CCITT).
     * If you change P, you must change crctab[]'s initial value to what is
     * printed by initcrctab()
     */
#define      P   0x1021

/* number of bits in CRC: don't change it. */
#define W    16

/* this the number of bits per char: don't change it. */
#define B    8

static unsigned short crctab[1<<B] = { /* as calculated by initcrctab() */
	0x0000,  0x1021,  0x2042,  0x3063,  0x4084,  0x50a5,  0x60c6,  0x70e7,
	0x8108,  0x9129,  0xa14a,  0xb16b,  0xc18c,  0xd1ad,  0xe1ce,  0xf1ef,
	0x1231,  0x0210,  0x3273,  0x2252,  0x52b5,  0x4294,  0x72f7,  0x62d6,
	0x9339,  0x8318,  0xb37b,  0xa35a,  0xd3bd,  0xc39c,  0xf3ff,  0xe3de,
	0x2462,  0x3443,  0x0420,  0x1401,  0x64e6,  0x74c7,  0x44a4,  0x5485,
	0xa56a,  0xb54b,  0x8528,  0x9509,  0xe5ee,  0xf5cf,  0xc5ac,  0xd58d,
	0x3653,  0x2672,  0x1611,  0x0630,  0x76d7,  0x66f6,  0x5695,  0x46b4,
	0xb75b,  0xa77a,  0x9719,  0x8738,  0xf7df,  0xe7fe,  0xd79d,  0xc7bc,
	0x48c4,  0x58e5,  0x6886,  0x78a7,  0x0840,  0x1861,  0x2802,  0x3823,
	0xc9cc,  0xd9ed,  0xe98e,  0xf9af,  0x8948,  0x9969,  0xa90a,  0xb92b,
	0x5af5,  0x4ad4,  0x7ab7,  0x6a96,  0x1a71,  0x0a50,  0x3a33,  0x2a12,
	0xdbfd,  0xcbdc,  0xfbbf,  0xeb9e,  0x9b79,  0x8b58,  0xbb3b,  0xab1a,
	0x6ca6,  0x7c87,  0x4ce4,  0x5cc5,  0x2c22,  0x3c03,  0x0c60,  0x1c41,
	0xedae,  0xfd8f,  0xcdec,  0xddcd,  0xad2a,  0xbd0b,  0x8d68,  0x9d49,
	0x7e97,  0x6eb6,  0x5ed5,  0x4ef4,  0x3e13,  0x2e32,  0x1e51,  0x0e70,
	0xff9f,  0xefbe,  0xdfdd,  0xcffc,  0xbf1b,  0xaf3a,  0x9f59,  0x8f78,
	0x9188,  0x81a9,  0xb1ca,  0xa1eb,  0xd10c,  0xc12d,  0xf14e,  0xe16f,
	0x1080,  0x00a1,  0x30c2,  0x20e3,  0x5004,  0x4025,  0x7046,  0x6067,
	0x83b9,  0x9398,  0xa3fb,  0xb3da,  0xc33d,  0xd31c,  0xe37f,  0xf35e,
	0x02b1,  0x1290,  0x22f3,  0x32d2,  0x4235,  0x5214,  0x6277,  0x7256,
	0xb5ea,  0xa5cb,  0x95a8,  0x8589,  0xf56e,  0xe54f,  0xd52c,  0xc50d,
	0x34e2,  0x24c3,  0x14a0,  0x0481,  0x7466,  0x6447,  0x5424,  0x4405,
	0xa7db,  0xb7fa,  0x8799,  0x97b8,  0xe75f,  0xf77e,  0xc71d,  0xd73c,
	0x26d3,  0x36f2,  0x0691,  0x16b0,  0x6657,  0x7676,  0x4615,  0x5634,
	0xd94c,  0xc96d,  0xf90e,  0xe92f,  0x99c8,  0x89e9,  0xb98a,  0xa9ab,
	0x5844,  0x4865,  0x7806,  0x6827,  0x18c0,  0x08e1,  0x3882,  0x28a3,
	0xcb7d,  0xdb5c,  0xeb3f,  0xfb1e,  0x8bf9,  0x9bd8,  0xabbb,  0xbb9a,
	0x4a75,  0x5a54,  0x6a37,  0x7a16,  0x0af1,  0x1ad0,  0x2ab3,  0x3a92,
	0xfd2e,  0xed0f,  0xdd6c,  0xcd4d,  0xbdaa,  0xad8b,  0x9de8,  0x8dc9,
	0x7c26,  0x6c07,  0x5c64,  0x4c45,  0x3ca2,  0x2c83,  0x1ce0,  0x0cc1,
	0xef1f,  0xff3e,  0xcf5d,  0xdf7c,  0xaf9b,  0xbfba,  0x8fd9,  0x9ff8,
	0x6e17,  0x7e36,  0x4e55,  0x5e74,  0x2e93,  0x3eb2,  0x0ed1,  0x1ef0
};


unsigned short
updcrc(unsigned short icrc, unsigned char *icp, unsigned int icnt) {
	register unsigned short crc = icrc;
	register unsigned char *cp = icp;
	register unsigned int cnt = icnt;

	while ( cnt-- ) {
		crc = (crc<<B) ^ crctab[(crc>>(W-B)) ^ *cp++];
	}

	return( crc );
}

#ifdef MAKETAB
main()
{
	initcrctab();
}

initcrctab()
{
	register  b, v, i;


	for ( b = 0; b <= (1<<B)-1; ++b ) {
		for ( v = b<<(W-B), i = B; --i >= 0; )
			v = v&0x8000 ? (v<<1) ^P : v<<1;
		crctab[b] = v;

		printf( "0x%04x,", v & 0xFFFF );
		if ( (b&7) == 7 )
			printf("\n" );
		else
			printf("  ");
	}
}
#endif

#ifdef TEST

#include <stdio.h>
#include <fcntl.h>

#define MAXBUF  32000

main( ac, av )
int ac; 
char **av;
{
	int fd = 0;
	int nr;
	char buf[MAXBUF];
	unsigned short crc;

	if ( ac > 1 )
		if ( (fd = open( av[1], O_RDONLY )) < 0 ) {
			perror( av[1] );
			exit( -1 );
		}
	crc = 0;
	while ( (nr = read( fd, buf, MAXBUF )) > 0 )
		crc = updcrc( crc, buf, nr );
	printf( "%04x\n", crc );
	if ( nr != 0 )
		perror( "reading" );
}

#endif

